/*
Copyright (c) 2014 by Şükrü Çınar.

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright notice,
   this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This program contains some code from the BSD kernel, which is copyrighted
by the Regents of the University of California. Their copyright statement
and software license is as follows:

 Copyright (c) 1986, 1988, 1991, 1993
	The Regents of the University of California.  All rights reserved.
 (c) UNIX System Laboratories, Inc.  All or some portions of this file
 are derived from material licensed to the University of California
 by American Telephone and Telegraph Co. or Unix System Laboratories,
 Inc. and are reproduced herein with the permission of UNIX System
 Laboratories, Inc.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met: 
 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 4. Neither the name of the University nor the names of its contributors
    may be used to endorse or promote products derived from this software
    without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/


/* This function does the following transformations:

    A/B/../C    -> A/C
    /./         -> /
    //          -> /
    ./A         -> A
*/
void fnmani_normalize(char *name)
{
   int len;
   int i,j;
   char *F,*G,*psearch;

   len= strlen(name);
again:
   if (name[0]==0) return ;
   if (name[0]=='.' && name[1]=='/') name[0]= name[1]= 0; 
   else if ((F= strstr(name,"//")))  F[0]= 0; 
   else if ((F= strstr(name,"/./"))) F[0]= F[1]= 0; 
   else goto parent;

   goto compact;
parent:

   psearch= name;

next_parent:

   F= strstr(psearch,"/../");
   if (!F) return ;
   if (F==name) { F[0]= F[1]= F[2]= 0; goto compact; }

   G= F-1;
   while(G>=name && G[0]!='/') G--;
   if (G<name || G[0]!='/') { psearch= F+1; goto next_parent; }

   if (G[1]=='.' && G[2]=='.' && G+3==F) { psearch= F+1; goto next_parent; }

   while(G<F+3) { G[0]= 0; G++; }

compact:
   j= 0;
   for(i=0;i<len;i++) if (name[i]) name[j++]= name[i];
   name[j]= 0; len= j;
   goto again;
}

/* returns the directory of the path. ignores any trailing / characters */
char *fnmani_dir(char *path)
{
   char* E;
   path= strdup(path);
   fnmani_normalize(path);
   for(E=path+strlen(path)-2;E>=path;E--)
      if (E[0]=='/') break;
   if (E>path) { E[0]= 0; return path; }
   free(path);
   if (E<path) return strdup(""); 
   return strdup("/");
}

/* This function returns a sibling of the file referenced by 'ref'.
   Both ref and sib can contain directory names, here's how everything
   is done:

   let be:
        D(ref)   directory part of ref
        D(sib)   directory part of sib

   if D(sib) is absolute, return sib
   return D(ref) concat sib

   The resulting string is normalized, by processing stuff such as
   /../ etc. */
char *fnmani_sibling(char *ref, char *sib)
{
   char *Dref,*Dsib;
   char *R;

   Dsib= fnmani_dir(sib);
   if (Dsib[0]=='/') { free(Dsib); return strdup(sib); }
   free(Dsib);

   Dref= fnmani_dir(ref);
   if (Dref[0]==0) { free(Dref); return strdup(sib); }
   R= malloc(strlen(Dref)+1+strlen(sib)+1);
   sprintf(R, "%s/%s", Dref, sib);
   free(Dref);
   
   fnmani_normalize(R);
   return R;
}

/* We are assuming that 'parent' is a directory name. There is only
   one special case here, parent can be completely empty. This means
   that  it was obtained from dir() and there was no directory part
   in the argument to dir(). In this case, we simply return the child
   name directly. */
char *fnmani_child(char *parent, char *name)
{
   char *K;
   if (parent[0]==0) return strdup(name);
   K= malloc(strlen(parent)+1 +strlen(name)+1);
   sprintf(K, "%s/%s", parent, name);
   fnmani_normalize(K);
   return K;
}

char *fnmani_file(char *path)
{
   char *K,*R;
   K= path+strlen(path)-1;

   // skip trailing /
   while (K>=path && K[0]=='/') K--;
   // this was all ///////
   if (K<path) { return strdup(""); }

   while(K>=path && K[0]!='/') K--;
   // no leading / was found.
   if (K<path) K= path;
   else K++;
   R= strdup(K);
   K= R+strlen(R)-1;
   while(K>=R && K[0]=='/') { K[0]= 0; K--; }
   return R;
}
